from redash.query_runner import query_runners
from redash_stmo.query_runner.big_query import BigQuery, parse_annotated_query
from tests import BaseTestCase


class TestBigQueryRunner(BaseTestCase):
    def test_custom_big_query_query_runner_registered(self):
        assert query_runners["bigquery"] == BigQuery

    def test_custom_big_query_query_runner_get_job_data(self, *args, **kwargs):
        query = """
            /* Task ID: 6ff0c989b3ff499193ca3fbb54225f80, Query Hash: 56888d88acb5475b95484eeb005f8e5f, Query ID: 12345, Queue: default, Username: jezdez */

            SELECT * FROM users;
        """
        with self.app.test_request_context("/"):
            self.app.preprocess_request()
            query_runner = BigQuery({})
            job_data = query_runner._get_job_data(query)
            assert job_data["labels"] == {
                "App": "redash",
                "Name": "localhost:5000/queries/12345",
                "Query Hash": "56888d88acb5475b95484eeb005f8e5f",
                "Query ID": "12345",
                "Queue": "default",
                "Task ID": "6ff0c989b3ff499193ca3fbb54225f80",
                "Username": "jezdez",
            }

    def test_parse_annotated_query_success(self):
        query = """
            /* Task ID: 5b7efa616d474a54874d337cc27f1953, Query Hash: 93ff42fc6ac443bfba222553890b4124, Query ID: 12345, Queue: default, Username: jezdez */

            SELECT * FROM users;
        """
        result = {
            "Query Hash": "93ff42fc6ac443bfba222553890b4124",
            "Query ID": "12345",
            "Queue": "default",
            "Task ID": "5b7efa616d474a54874d337cc27f1953",
            "Username": "jezdez",
        }
        assert parse_annotated_query(query) == result

    def test_parse_annotated_query_no_annotation(self):
        assert parse_annotated_query("") == {}
        assert parse_annotated_query("SELECT * FROM users;") == {}

    def test_parse_annotated_query_wrong_annotation(self):
        query = """
            /* Task ID: 8ccd40c878f59fa69ccf31a72140b208, Query Hash: f6bf37efedbc0a2dfffc1caf5088d86e, Query ID: 12345, Queue: default, Username: jezdez

            SELECT * FROM users;
        """
        result = {}
        assert parse_annotated_query(query) == result



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, you can obtain one at http://mozilla.org/MPL/2.0/.
import json
import os
import time
from datetime import timedelta

from flask import jsonify
from flask_login import login_required
from redash import models, redis_connection, statsd_client
from redash.handlers.base import BaseResource, routes
from redash.monitor import get_status as original_get_status
from redash.permissions import require_super_admin
from redash.utils import parse_human_time
from redash.worker import get_job_logger, job
from redash_stmo import settings
from redash_stmo.resources import add_resource

logger = get_job_logger(__name__)


class DataSourceHealthResource(BaseResource):
    def get(self):
        health_data = json.loads(redis_connection.get("data_sources:health") or "{}")
        return jsonify(health_data)


def store_health_status(data_source_id, data_source_name, query_text, data):
    key = "data_sources:health"

    cache = json.loads(redis_connection.get(key) or "{}")
    if data_source_id not in cache:
        cache[data_source_id] = {"metadata": {"name": data_source_name}, "queries": {}}
    cache[data_source_id]["queries"][query_text] = data

    cache[data_source_id]["status"] = "SUCCESS"
    for query_status in cache[data_source_id]["queries"].values():
        if query_status["status"] == "FAIL":
            cache[data_source_id]["status"] = "FAIL"
            break

    redis_connection.set(key, json.dumps(cache))


@job("default", timeout=settings.HEALTH_QUERIES_REFRESH_TIMEOUT)
def update_health_status():
    for data_source in models.DataSource.query:
        logger.info(u"job=update_health_status state=start ds_id=%s", data_source.id)

        runtime = None
        query_text = data_source.query_runner.noop_query
        ds_id = str(data_source.id)

        custom_query_env_var = "REDASH_CUSTOM_HEALTH_QUERIES_{data_source_id}".format(
            data_source_id=ds_id
        )
        custom_query = os.environ.get(custom_query_env_var, "")
        query_text = custom_query or query_text

        try:
            start_time = time.time()
            test_connection(data_source.query_runner, query_text)
            runtime = time.time() - start_time
        except NotImplementedError:
            logger.info(
                u"Unable to compute health status without test query for %s",
                data_source.name,
            )
            continue
        except Exception:
            logger.warning(
                u"Failed health check for the data source: %s",
                data_source.name,
                exc_info=1,
            )
            statsd_client.incr("update_health_status.error")
            logger.info(
                u"job=update_health_status state=error ds_id=%s runtime=%.2f",
                data_source.id,
                time.time() - start_time,
            )

        status = {
            "status": "FAIL" if runtime is None else "SUCCESS",
            "last_run": start_time,
            "last_run_human": str(parse_human_time(str(start_time))),
            "runtime": runtime,
        }
        store_health_status(ds_id, data_source.name, query_text, status)


def test_connection(self, custom_query_text=None):
    if self.noop_query is None:
        raise NotImplementedError()

    query_text = custom_query_text or self.noop_query
    data, error = self.run_query(query_text, None)

    if error is not None:
        raise Exception(error)


@login_required
@require_super_admin
def stmo_status_api():
    status = original_get_status()
    health_data = json.loads(redis_connection.get("data_sources:health") or "{}")

    # Get the top level status for each data source
    for health_data_point in health_data.values():
        data_source_name = health_data_point["metadata"]["name"]
        dashboard_label = "[Data Source Health] {name}".format(name=data_source_name)
        status[dashboard_label] = health_data_point["status"]
    return jsonify(status)


def extension(app=None):
    """A Redash extension to add datasource health status reporting."""

    # Override the default status API view with our extended view
    app.view_functions["%s.status_api" % routes.name] = stmo_status_api

    # Add a new endpoint with full health data
    add_resource(app, DataSourceHealthResource, "/status/data_sources/health.json")


def scheduled_job():
    """Add the update_health_status task to a list of scheduled jobs"""
    return {
        "func": update_health_status,
        "interval": timedelta(minutes=settings.HEALTH_QUERIES_REFRESH_SCHEDULE),
    }



from redash.models import Group, db
from tests import BaseTestCase, authenticate_request

from redash_stmo.handlers.query_results.parser import extract_table_names


class TestQueryResultAPI(BaseTestCase):
    def test_query_result_query(self):
        restricted_group = self.factory.create_group(
            org=self.factory.org, type=Group.REGULAR_GROUP, name="restricted"
        )
        # Commit needed because User.group_ids is an array of IDs, not a relation
        db.session.commit()
        restricted_user = self.factory.create_user(
            group_ids=[restricted_group.id, self.factory.org.default_group.id],
            name="Restricted User",
        )
        restricted_ds = self.factory.create_data_source(
            group=restricted_group, view_only=False, type="pg"
        )
        upstream_query = self.factory.create_query(
            user=restricted_user,
            data_source=restricted_ds,
            query_text="select * from secret",
        )
        result_ds = self.factory.create_data_source(
            group=self.factory.org.default_group, view_only=False, type="results"
        )
        query = self.factory.create_query(
            user=restricted_user,
            data_source=result_ds,
            query_text="select password from query_{} limit 10".format(
                upstream_query.id
            ),
        )
        query_result = self.factory.create_query_result(
            data_source=result_ds,
            query_text=query.query_text,
            query_hash=query.query_hash,
        )
        authenticate_request(self.client, restricted_user)
        rv = self.client.get("/api/query_results/{}".format(query_result.id))
        self.assertEqual(rv.status_code, 200)

    def test_query_result_query_restrict_access(self):
        restricted_group = self.factory.create_group(
            org=self.factory.org, type=Group.REGULAR_GROUP, name="restricted"
        )
        restricted_user = self.factory.create_user(
            group_ids=[restricted_group.id, self.factory.org.default_group.id]
        )
        restricted_ds = self.factory.create_data_source(
            group=restricted_group, view_only=False, type="pg"
        )
        upstream_query = self.factory.create_query(
            user=restricted_user,
            data_source=restricted_ds,
            query_text="select * from secret",
        )
        result_ds = self.factory.create_data_source(
            group=self.factory.org.default_group, view_only=False, type="results"
        )
        query = self.factory.create_query(
            user=restricted_user,
            data_source=result_ds,
            query_text="select password from query_{}".format(upstream_query.id),
        )
        query_result = self.factory.create_query_result(
            data_source=result_ds,
            query_text=query.query_text,
            query_hash=query.query_hash,
        )
        authenticate_request(self.client, restricted_user)
        rv = self.client.get("/api/query_results/{}".format(query_result.id))
        self.assertEqual(rv.status_code, 403)

    def test_extract_table_names(self):
        query = """\
WITH standard AS (
    SELECT
        'Cold Startup from App Link' AS TestName,
        'fnprms' AS project,
        'Motorola G5' AS platform,
        NULL AS FennecValue,
        -20.0 AS Target,
        date AS TestGroupTime
    FROM cached_query_67650
    WHERE value > 0
        AND date >= datetime('now', '-56 day')
    ORDER BY TestGroupTime DESC
), fenix AS (
    SELECT
        standard.*,
        'fenix' AS BrowserTested,
        value AS CurrentValue,
        ((value - 0.511) / 0.511 * 100) AS CurrentPercent,
        (SELECT value FROM cached_query_67650 WHERE date <= datetime('now', '-7 day') ORDER BY date DESC LIMIT 1) AS LastWeekValue
    FROM cached_query_67650 AS full
        LEFT JOIN standard ON full.date = standard.TestGroupTime
    WHERE value > 0
), fennec AS (
    SELECT
        standard.*,
        'fennec' AS BrowserTested,
        standard.FennecValue AS CurrentValue,
        NULL AS CurrentPercent,
        0 AS LastWeekValue
    FROM fenix
        LEFT JOIN standard ON fenix.TestGroupTime = standard.TestGroupTime
), goal AS (
    SELECT
        standard.*,
        'target' AS BrowserTested,
        0.8 * standard.FennecValue AS CurrentValue,
        NULL AS CurrentPercent,
        0 AS LastWeekValue
    FROM fenix
        LEFT JOIN standard ON fenix.TestGroupTime = standard.TestGroupTime
),  combined AS (
    SELECT * FROM fenix
    UNION ALL
    SELECT * FROM fennec
    UNION ALL
    SELECT * FROM goal
)

SELECT
    *,
    (CurrentValue - LastWeekValue) / LastWeekValue * 100 AS OneWeekDifference
FROM combined
ORDER BY TestGroupTime DESC, BrowserTested

"""
        self.assertEqual(extract_table_names(query), ["combined"])



import mock
import json

from redash_stmo.data_sources.details.extension import DATASOURCE_URLS
from redash.utils.configuration import ConfigurationContainer
from tests import BaseTestCase, authenticate_request


class TestDataSourceDetails(BaseTestCase):
    EXPECTED_DOC_URL = "www.example.com"

    def setUp(self):
        super(TestDataSourceDetails, self).setUp()
        self.admin = self.factory.create_admin()
        self.data_source = self.factory.create_data_source(
            options=lambda: ConfigurationContainer.from_json(
                json.dumps(
                    dict(
                        user="postgres",
                        password="postgres",
                        host="postgres",
                        dbname="tests",
                    )
                )
            ),
        )
        self.patched_run_query = self._setup_mock(
            "redash.query_runner.pg.PostgreSQL.run_query"
        )
        self.patched_runner_type = self._setup_mock(
            "redash.query_runner.pg.PostgreSQL.type"
        )
        authenticate_request(self.client, self.admin)

    def _setup_mock(self, function_to_patch):
        patcher = mock.patch(function_to_patch)
        patched_function = patcher.start()
        self.addCleanup(patcher.stop)
        return patched_function

    def _test_expected_version_returned(
        self, expected_version, version_string, runner_type
    ):
        self.patched_runner_type.return_value = runner_type
        self.patched_run_query.return_value = (
            json.dumps(
                {"rows": [{"version": version_string.format(version=expected_version)}]}
            ),
            None,
        )
        rv = self.client.get(
            "/api/data_sources/{}/details".format(self.data_source.id),
        )
        self.assertEqual(200, rv.status_code)
        self.assertEqual(json.loads(rv.data)["message"]["version"], expected_version)

    def test_gets_postgres_version(self):
        RUNNER_TYPE = "pg"
        DATASOURCE_VERSION = "9.5.10"
        VERSION_STRING = (
            "PostgreSQL {version} on x86_64-pc-linux-gnu, compiled by gcc "
            "(GCC) 4.8.3 20140911 (Red Hat 4.8.3-9), 64-bit"
        )
        self._test_expected_version_returned(
            DATASOURCE_VERSION, VERSION_STRING, RUNNER_TYPE
        )

    def test_gets_redshift_version(self):
        RUNNER_TYPE = "redshift"
        DATASOURCE_VERSION = "1.0.3688"
        VERSION_STRING = (
            "PostgreSQL 8.0.2 on i686-pc-linux-gnu, compiled by GCC "
            "gcc (GCC) 3.4.2 20041017 (Red Hat 3.4.2-6.fc3), Redshift {version}"
        )
        self._test_expected_version_returned(
            DATASOURCE_VERSION, VERSION_STRING, RUNNER_TYPE
        )

    def test_gets_mysql_version(self):
        RUNNER_TYPE = "mysql"
        DATASOURCE_VERSION = "5.7.16"
        VERSION_STRING = "{version}-log"
        self._test_expected_version_returned(
            DATASOURCE_VERSION, VERSION_STRING, RUNNER_TYPE
        )

    def test_unexpected_json(self):
        self.patched_runner_type.return_value = "pg"
        self.patched_run_query.return_value = (
            json.dumps({"rows": [{"bad_json": "foo"}]}),
            None,
        )
        rv = self.client.get(
            "/api/data_sources/{}/details".format(self.data_source.id),
        )
        self.assertEqual(200, rv.status_code)
        self.assertEqual(json.loads(rv.data)["message"]["version"], None)

    def test_gets_datasource_link_and_type(self):
        self.patched_runner_type.return_value = "pg"
        VERSION_STRING = (
            "PostgreSQL 8.0.2 on i686-pc-linux-gnu, compiled by GCC "
            "gcc (GCC) 3.4.2 20041017 (Red Hat 3.4.2-6.fc3), Redshift {version}"
        )
        self.patched_run_query.return_value = (
            json.dumps({"rows": [{"version": VERSION_STRING}]}),
            None,
        )
        rv = self.client.get("/api/data_sources/{}/details".format(self.data_source.id))
        self.assertEqual(200, rv.status_code)
        rv_json = json.loads(rv.data)
        self.assertEqual(
            rv_json["message"]["type_name"], self.data_source.query_runner.name()
        )
        self.assertEqual(rv_json["message"]["doc_url"], DATASOURCE_URLS["pg"])



import logging
import re

from redash.authentication.org_resolving import current_org
from redash.query_runner import register
from redash.query_runner.big_query import BigQuery as RedashBigQuery
from redash.utils import base_url

logger = logging.getLogger(__name__)


ANNOTATION_RE = re.compile(r"^\/\*\s(.*)\s\*\/$", re.U | re.M)


def parse_annotated_query(query):
    """
    Parses the given query for the annotation that Redash left
    there when before running the job.

    E.g. a query that has the annotation on top::

        /* Task ID: 8ccd40c878f59fa69ccf31a72140b208, Query Hash: f6bf37efedbc0a2dfffc1caf5088d86e, Query ID: 12345, Queue: default, Username: jezdez */

        SELECT * FROM users;

    will lead to returning::

        {
            'Query Hash': 'f6bf37efedbc0a2dfffc1caf5088d86e',
            'Query ID': '12345',
            'Queue': 'default',
            'Task ID': '8ccd40c878f59fa69ccf31a72140b208',
            'Username': 'jezdez',
        }

    which we can use as labels when submitting the BigQuery job.
    """
    if not query or "/*" not in query:
        return {}
    match = ANNOTATION_RE.match(query.strip())
    if not match:
        return {}
    # Split by comma and colons to create a key/value dict of query annotations
    return dict(
        map(lambda s: (x.strip() for x in s.split(":")), match.group(1).split(","))
    )


class BigQuery(RedashBigQuery):
    # Needed so we can extract annotations from query for job labels
    should_annotate_query = True

    @classmethod
    def type(cls):
        """Overrides the name to match the name of the parent query runner"""
        return "bigquery"

    def __init__(self, *args, **kwargs):
        super(BigQuery, self).__init__(*args, **kwargs)
        self._query_user = None

    def _get_job_data(self, query):
        job_data = super(BigQuery, self)._get_job_data(query)
        labels = {"App": "redash"}

        # Add the Owner label with the current user's email address
        if self._query_user is not None:
            labels["Owner"] = self._query_user.email

        # Add all the parsed query metadata as labels to the job
        parsed_annotation = parse_annotated_query(query)
        labels.update(parsed_annotation)

        # Add a full URL to the query for the "Name" label
        if "Query ID" in parsed_annotation:
            host = base_url(current_org)
            labels["Name"] = "{host}/queries/{query_id}".format(
                host=host, query_id=parsed_annotation["Query ID"]
            )

        if labels:
            job_data["labels"] = labels

        return job_data

    def run_query(self, query, user):
        # Storing the query's user in an instance variable for use
        # in the custom job label.
        self._query_user = user
        return super(BigQuery, self).run_query(query, user)


def extension(app):
    logger.info("Loading Redash Extension for the custom BigQuery query runner")
    # Register our own BigQuery query runner class
    # which automatically overwrites the default presto query runner
    register(BigQuery)
    logger.info("Loaded Redash Extension for the custom BigQuery query runner")
    return



import os.path

from setuptools import find_packages, setup

readme = ""
here = os.path.abspath(os.path.dirname(__file__))
readme_path = os.path.join(here, "README.rst")
if os.path.exists(readme_path):
    with open(readme_path, "rb") as stream:
        readme = stream.read().decode("utf8")


setup(
    long_description=readme,
    long_description_content_type="text/x-rst",
    name="redash-stmo",
    use_scm_version={"version_scheme": "post-release", "local_scheme": "dirty-tag"},
    setup_requires=["setuptools_scm"],
    description="Extensions to Redash by Mozilla",
    project_urls={"homepage": "https://github.com/mozilla/redash-stmo"},
    author="Mozilla Foundation",
    author_email="dev-webdev@lists.mozilla.org",
    license="MPL-2.0",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Web Environment :: Mozilla",
        "Framework :: Django",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Topic :: Internet :: WWW/HTTP",
    ],
    entry_points={
        "redash.extensions": [
            "datasource_health = redash_stmo.data_sources.health:extension",
            "datasource_details = redash_stmo.data_sources.details.extension:extension",
            "dockerflow = redash_stmo.dockerflow:extension",
            "handler_queryresults = redash_stmo.handlers.query_results.extension:extension",
            "handler_remote_user_auth = redash_stmo.handlers.authentication.remote_user_auth:extension",
            "queryrunner_big_query = redash_stmo.query_runner.big_query:extension",
            "queryrunner_presto = redash_stmo.query_runner.presto:extension",
        ],
        "redash.bundles": [
            "datasource_details = redash_stmo.data_sources.details",
        ],
        "redash.scheduled_jobs": [
            "update_health_status = redash_stmo.data_sources.health:scheduled_job"
        ],
    },
    packages=find_packages("src"),
    package_dir={"": "src"},
    include_package_data=True,
    python_requires='>=3.5, <4',
    install_requires=[
        "dockerflow>=2018.4.0",
        "pyhive",
        "requests",
        "sqlparse",
    ],
    extras_require={
        "test": [
            "flake8==3.5.0",
            "mock",
            "pytest",
            "pytest-cov",
            "pytest-flake8<1.0.1",
        ],
        "dev": ["watchdog"],
    },
)
